import numpy

from typing import List
from suber.data_types import Segment, Subtitle


def time_align_hypothesis_to_reference(hypothesis: List[Segment], reference: List[Subtitle]) -> List[Subtitle]:
    """
    Re-segments the hypothesis segments according to the reference subtitle timings. The output hypothesis subtitles
    will have the same time stamps as the reference, and each will contain the words whose approximate times falls into
    these intervals, i.e. reference_subtitle.start_time < word.approximate_word_time < reference_subtitle.end_time.
    Hypothesis words that do not fall into any subtitle will be dropped.
    """
    aligned_hypothesis_word_lists = [[] for _ in reference]

    reference_start_times = numpy.array([subtitle.start_time for subtitle in reference])
    reference_end_times = numpy.array([subtitle.end_time for subtitle in reference])

    for segment in hypothesis:
        for word in segment.word_list:
            assert word.approximate_word_time is not None, "Should have been set by SRTFileReader. Is plain file used?"
            reference_subtitle_index = numpy.searchsorted(reference_start_times, word.approximate_word_time) - 1

            if reference_subtitle_index < 0:
                # Word is before first subtitle, drop it.
                continue

            if word.approximate_word_time < reference_end_times[reference_subtitle_index]:
                aligned_hypothesis_word_lists[reference_subtitle_index].append(word)

    aligned_hypothesis = []

    for index, word_list in enumerate(aligned_hypothesis_word_lists):
        reference_subtitle = reference[index]
        subtitle = Subtitle(
            word_list=word_list,
            index=reference_subtitle.index,
            start_time=reference_subtitle.start_time,
            end_time=reference_subtitle.end_time)

        aligned_hypothesis.append(subtitle)

    return aligned_hypothesis



#!/usr/bin/env python3

import argparse

from suber.__main__ import read_input_file
from suber.hyp_to_ref_alignment import levenshtein_align_hypothesis_to_reference
from suber.hyp_to_ref_alignment import time_align_hypothesis_to_reference
from suber.utilities import segment_to_string


def parse_arguments():
    parser = argparse.ArgumentParser(description="Re-segments the hypothesis file to match the reference. This can "
                                                 "either be done via Levenshtein alignment, or using the subtitle "
                                                 "timings, if available.")
    parser.add_argument("-H", "--hypothesis", required=True, help="The input file.")
    parser.add_argument("-R", "--reference", required=True, help="The reference file.")
    parser.add_argument("-o", "--aligned-hypothesis", required=True,
                        help="The aligned hypothesis output file in plain format.")
    parser.add_argument("-f", "--hypothesis-format", default="SRT", help="Hypothesis file format, 'SRT' or 'plain'.")
    parser.add_argument("-F", "--reference-format", default="SRT", help="Reference file format, 'SRT' or 'plain'.")
    parser.add_argument("-m", "--method", default="levenshtein",
                        help="The alignment method, either 'levenshtein' or 'time'. See the "
                             "'suber.hyp_to_ref_alignment' module. 'time' only supported if both hypothesis and "
                             "reference are given in SRT format.")

    return parser.parse_args()


def main():
    args = parse_arguments()

    if args.method == "time" and not args.hypothesis_format == "SRT" and args.reference_format == "SRT":
        raise ValueError("For time alignment, both hypothesis and reference have to be given in SRT format.")

    hypothesis_segments = read_input_file(args.hypothesis, file_format=args.hypothesis_format)
    reference_segments = read_input_file(args.reference, file_format=args.reference_format)

    if args.method == "levenshtein":
        aligned_hypothesis_segments = levenshtein_align_hypothesis_to_reference(
            hypothesis=hypothesis_segments, reference=reference_segments)
    elif args.method == "time":
        aligned_hypothesis_segments = time_align_hypothesis_to_reference(
            hypothesis=hypothesis_segments, reference=reference_segments)

    with open(args.aligned_hypothesis, "w") as output_file_object:
        for segment in aligned_hypothesis_segments:
            output_file_object.write(segment_to_string(segment) + '\n')


if __name__ == "__main__":
    main()



import numpy
import string
import Levenshtein
from itertools import zip_longest
from typing import List, Tuple

from suber.data_types import Segment


def levenshtein_align_hypothesis_to_reference(hypothesis: List[Segment], reference: List[Segment]) -> List[Segment]:
    """
    Runs the Levenshtein algorithm to get the minimal set of edit operations to convert the full list of hypothesis
    words into the full list of reference words. The edit operations implicitly define an alignment between hypothesis
    and reference words. Using this alignment, the hypotheses are re-segmented to match the reference segmentation.
    """

    remove_punctuation_table = str.maketrans('', '', string.punctuation)

    def normalize_word(word):
        """
        Lower-cases and removes punctuation as this increases the alignment accuracy.
        """
        word = word.lower()
        word_without_punctuation = word.translate(remove_punctuation_table)

        if not word_without_punctuation:
            return word  # keep tokens that are purely punctuation

        return word_without_punctuation

    all_reference_word_strings = [normalize_word(word.string) for segment in reference for word in segment.word_list]
    all_hypothesis_word_strings = [normalize_word(word.string) for segment in hypothesis for word in segment.word_list]

    all_hypothesis_words = [word for segment in hypothesis for word in segment.word_list]

    reference_string, hypothesis_string = _map_words_to_characters(
        all_reference_word_strings, all_hypothesis_word_strings)

    opcodes = Levenshtein.opcodes(reference_string, hypothesis_string)

    reference_segment_lengths = [len(segment.word_list) for segment in reference]
    reference_segment_boundary_indices = numpy.cumsum(reference_segment_lengths)
    current_segment_index = 0
    aligned_hypothesis_word_lists = [[] for _ in reference]

    for opcode_tuple in opcodes:
        edit_operation = opcode_tuple[0]
        hypothesis_position_range = range(opcode_tuple[3], opcode_tuple[4])
        reference_position_range = range(opcode_tuple[1], opcode_tuple[2])

        if edit_operation in ("equal", "replace"):
            assert len(hypothesis_position_range) == len(hypothesis_position_range)
        elif edit_operation == "insert":
            assert len(reference_position_range) == 0
        elif edit_operation == "delete":
            assert len(hypothesis_position_range) == 0
        else:
            assert False, f"Invalid edit operation '{edit_operation}'."

        # 'zip_longest' is a "clever" way to unify the different cases: for 'equal' and 'replace' we indeed have to
        # iterate through hypothesis and reference position in parallel, for 'insert' and 'delete' either
        # 'hypothesis_position' or 'reference_position' will be None in the loop.
        for hypothesis_position, reference_position in zip_longest(hypothesis_position_range, reference_position_range):

            # Update current segment index depending on current reference position.
            if (reference_position is not None
                    and reference_position >= reference_segment_boundary_indices[current_segment_index]):

                assert reference_position == reference_segment_boundary_indices[current_segment_index], (
                    "Bug: missing reference position in edit operations.")
                current_segment_index += 1

                # If there are empty segments in the reference, we get double entries in
                # 'reference_segment_boundary_indices' (because the empty segment ends at the same word index as the
                # previous segment). Skip these empty segments, we don't want to assign any hypothesis words to them.
                while (current_segment_index < len(reference_segment_boundary_indices)
                       and reference_segment_boundary_indices[current_segment_index]
                       == reference_segment_boundary_indices[current_segment_index - 1]):
                    current_segment_index += 1

            # Add hypothesis word to current segment in case of 'equal', 'replace' or 'insert' operation.
            if hypothesis_position is not None:
                word = all_hypothesis_words[hypothesis_position]
                aligned_hypothesis_word_lists[current_segment_index].append(word)

    aligned_hypothesis = [Segment(word_list=word_list) for word_list in aligned_hypothesis_word_lists]

    return aligned_hypothesis


def _map_words_to_characters(reference_words: List[str], hypothesis_words: List[str]) -> Tuple[str, str]:
    """
    The Levenshtein module operates on strings, not list of strings. Therefore we map words to characters here.
    Inspired by https://github.com/jitsi/jiwer/blob/master/jiwer/measures.py.
    """
    unique_words = set(reference_words + hypothesis_words)
    vocabulary = dict(zip(unique_words, range(len(unique_words))))

    reference_string = "".join(chr(vocabulary[word] + 32) for word in reference_words)
    hypothesis_string = "".join(chr(vocabulary[word] + 32) for word in hypothesis_words)

    return reference_string, hypothesis_string



from suber.data_types import LineBreak, Segment
from suber.constants import END_OF_LINE_SYMBOL, END_OF_BLOCK_SYMBOL, MASK_SYMBOL


def segment_to_string(segment: Segment, include_line_breaks=False, include_last_break=True,
                      mask_all_words=False) -> str:
    if not include_line_breaks:
        assert not mask_all_words, (
            "Refusing to mask all words when not printing breaks, output would contain only mask symbols.")
        return " ".join(word.string for word in segment.word_list)

    word_list_with_breaks = []
    for word in segment.word_list:
        word_list_with_breaks.append(MASK_SYMBOL if mask_all_words else word.string)

        if word.line_break == LineBreak.END_OF_LINE:
            word_list_with_breaks.append(END_OF_LINE_SYMBOL)
        elif word.line_break == LineBreak.END_OF_BLOCK:
            word_list_with_breaks.append(END_OF_BLOCK_SYMBOL)

    if not include_last_break and word_list_with_breaks and word_list_with_breaks[-1] == END_OF_BLOCK_SYMBOL:
        word_list_with_breaks.pop()

    return " ".join(word_list_with_breaks)


def get_segment_to_string_opts_from_metric(metric: str):
    include_breaks = False
    mask_words = False
    if metric.endswith("-br"):
        include_breaks = True
        mask_words = True
        metric = metric[:-len("-br")]
    elif metric.endswith("-seg"):
        include_breaks = True
        metric = metric[:-len("-seg")]

    return include_breaks, mask_words, metric



import jiwer
import functools
from typing import List

from suber.data_types import Segment
from suber.utilities import segment_to_string, get_segment_to_string_opts_from_metric


def calculate_word_error_rate(hypothesis: List[Segment], reference: List[Segment], metric="WER",
                              score_break_at_segment_end=True) -> float:

    assert len(hypothesis) == len(reference), (
        "Number of hypothesis segments does not match reference, alignment step missing?")

    if metric == "WER-cased":
        transformations = jiwer.Compose([
            jiwer.ReduceToListOfListOfWords(),
        ])
        metric = "WER"

    else:
        transformations = jiwer.Compose([
            jiwer.ToLowerCase(),
            jiwer.RemovePunctuation(),
            # Ellipsis is a common character in subtitles that jiwer would not remove by default.
            jiwer.RemoveSpecificWords(['…']),
            jiwer.ReduceToListOfListOfWords(),
        ])

    include_breaks, mask_words, metric = get_segment_to_string_opts_from_metric(metric)
    assert metric == "WER"

    segment_to_string_ = functools.partial(
        segment_to_string, include_line_breaks=include_breaks, mask_all_words=mask_words,
        include_last_break=score_break_at_segment_end)

    hypothesis_strings = [segment_to_string_(segment) for segment in hypothesis]
    reference_strings = [segment_to_string_(segment) for segment in reference]

    wer_score = jiwer.wer(
        reference_strings, hypothesis_strings, truth_transform=transformations, hypothesis_transform=transformations)

    return round(wer_score * 100, 3)



import functools
from typing import List

from sacrebleu.metrics import BLEU, TER, CHRF

from suber.data_types import Segment
from suber.utilities import segment_to_string, get_segment_to_string_opts_from_metric


def calculate_sacrebleu_metric(hypothesis: List[Segment], reference: List[Segment],
                               metric="BLEU", score_break_at_segment_end=True) -> float:

    assert len(hypothesis) == len(reference), (
        "Number of hypothesis segments does not match reference, alignment step missing?")

    include_breaks, mask_words, metric = get_segment_to_string_opts_from_metric(metric)

    if metric == "BLEU":
        sacrebleu_metric = BLEU()
    elif metric == "TER":
        sacrebleu_metric = TER()
    elif metric == "chrF":
        sacrebleu_metric = CHRF()
    else:
        raise ValueError(f"Unsupported sacrebleu metric '{metric}'.")

    # Sacrebleu currently does not allow empty references, just skip empty reference segments as a workaround.
    if not all(segment.word_list for segment in reference):
        empty_reference_indices = [index for index, segment in enumerate(reference) if not segment.word_list]
        reference = [segment for index, segment in enumerate(reference) if index not in empty_reference_indices]
        hypothesis = [segment for index, segment in enumerate(hypothesis) if index not in empty_reference_indices]

    segment_to_string_ = functools.partial(
        segment_to_string, include_line_breaks=include_breaks, mask_all_words=mask_words,
        include_last_break=score_break_at_segment_end)

    hypothesis_strings = [segment_to_string_(segment) for segment in hypothesis]
    reference_strings = [[segment_to_string_(segment) for segment in reference]]  # sacrebleu expects nested list

    if include_breaks:
        # BLEU tokenizer would split "<eol>" into "< eol >".
        hypothesis_strings = [string.replace("<eol>", "eol").replace("<eob>", "eob") for string in hypothesis_strings]
        reference_strings[0] = [
            string.replace("<eol>", "eol").replace("<eob>", "eob") for string in reference_strings[0]]

    sacrebleu_score = sacrebleu_metric.corpus_score(hypotheses=hypothesis_strings, references=reference_strings)

    return round(sacrebleu_score.score, 3)



import Levenshtein
import string
from typing import List

from suber.data_types import Segment
from suber.utilities import segment_to_string


def calculate_character_error_rate(hypothesis: List[Segment], reference: List[Segment], metric="CER") -> float:
    assert len(hypothesis) == len(reference), (
        "Number of hypothesis segments does not match reference, alignment step missing?")

    hypothesis_strings = [segment_to_string(segment) for segment in hypothesis]
    reference_strings = [segment_to_string(segment) for segment in reference]

    if metric != "CER-cased":
        remove_punctuation_table = str.maketrans('', '', string.punctuation)

        def normalize_string(string):
            string = string.translate(remove_punctuation_table)
            # Ellipsis is a common character in subtitles which is not included in string.punctuation.
            string = string.replace('…', '')
            string = string.lower()
            return string

        hypothesis_strings = [normalize_string(string) for string in hypothesis_strings]
        reference_strings = [normalize_string(string) for string in reference_strings]

    num_edits = 0
    num_reference_characters = 0
    for hypothesis_string, reference_string, in zip(hypothesis_strings, reference_strings):
        num_edits += Levenshtein.distance(hypothesis_string, reference_string)
        num_reference_characters += len(reference_string)

    cer_score = num_edits / num_reference_characters

    return round(cer_score * 100, 3)



import gzip

from typing import List
from io import TextIOWrapper

from suber.data_types import Segment


class FileReaderBase:
    """
    Derived classes must implement self._parse_lines().
    """
    def __init__(self, file_name):
        self._file_name = file_name

    def read(self) -> List[Segment]:
        with self._open_file() as file_object:
            return list(self._parse_lines(file_object))

    def _parse_lines(self, file_object: TextIOWrapper) -> List[Segment]:
        raise NotImplementedError

    def _open_file(self):
        if self._file_name.endswith(".gz"):
            return gzip.open(self._file_name, 'rt')
        else:
            return open(self._file_name, 'r')



from typing import List
from suber.data_types import Segment


sentence_final_punctuation = ['.', '?', '!', '！', '？', '。', "…"]
quotation_marks = ["'", '"']
# all combinations of punctuation and quotes, i.e. '."', "?'" etc.
quoted_sentence_final_punctuation = [punct + quote for punct in sentence_final_punctuation for quote in quotation_marks]
ellipses = ["...", "…"]


def resegment_based_on_punctuation(segments: List[Segment]) -> List[Segment]:
    resegmented_segments = []

    all_words = [word for segment in segments for word in segment.word_list]

    word_list = []
    previous_word = None
    for word in all_words:
        if not previous_word or _is_sentence_end(previous_word.string, word.string):
            if word_list:
                resegmented_segments.append(Segment(word_list=word_list))
            word_list = [word]
        else:
            word_list.append(word)
        previous_word = word

    assert word_list
    resegmented_segments.append(Segment(word_list=word_list))

    return resegmented_segments


def _is_sentence_end(current_word: str, next_word: str = None):
    if not next_word:
        # No next word, force sentence end.
        return True

    assert current_word, "'current_word' must not be empty."

    is_sentence_final_punctuation_at_end = (
        current_word[-1] in sentence_final_punctuation
        or (len(current_word) > 1 and current_word[-2:] in quoted_sentence_final_punctuation))

    is_ellipsis_at_end = any(current_word.endswith(ellipsis) for ellipsis in ellipses)

    next_word_is_lower_cased = next_word[0].islower()
    next_word_is_lower_or_digit = next_word_is_lower_cased or next_word[0].isdigit()

    is_sentence_end = (is_sentence_final_punctuation_at_end and not next_word_is_lower_cased
                       and not (is_ellipsis_at_end and next_word_is_lower_or_digit))

    return is_sentence_end



#!/usr/bin/env python3

import argparse
import json

from collections import OrderedDict

from suber.file_readers import PlainFileReader, SRTFileReader
from suber.hyp_to_ref_alignment import levenshtein_align_hypothesis_to_reference
from suber.hyp_to_ref_alignment import time_align_hypothesis_to_reference
from suber.metrics.suber import calculate_SubER
from suber.metrics.sacrebleu_interface import calculate_sacrebleu_metric
from suber.metrics.jiwer_interface import calculate_word_error_rate
from suber.metrics.cer import calculate_character_error_rate


def parse_arguments():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="SubER - Subtitle Edit Rate. An automatic, reference-based, segmentation- and timing-aware "
                    "edit distance metric to measure quality of subtitle files. Basic usage: "
                    "'python -m suber -H hypothesis.srt -R reference.srt'")
    parser.add_argument("-H", "--hypothesis", required=True, help="The input file to score.")
    parser.add_argument("-R", "--reference", required=True, help="The reference file.")
    parser.add_argument("-m", "--metrics", nargs="+", default=["SubER"], help="The metrics to compute.")
    parser.add_argument("-f", "--hypothesis-format", default="SRT", help="Hypothesis file format, 'SRT' or 'plain'.")
    parser.add_argument("-F", "--reference-format", default="SRT", help="Reference file format, 'SRT' or 'plain'.")

    return parser.parse_args()


def main():
    args = parse_arguments()

    check_metrics(args.metrics)
    check_file_formats(args.hypothesis_format, args.reference_format, args.metrics)

    # A "segment" is a subtitle in case of SRT file input, or a line of text in case of plain input.
    hypothesis_segments = read_input_file(args.hypothesis, file_format=args.hypothesis_format)
    reference_segments = read_input_file(args.reference, file_format=args.reference_format)

    # Aligned hypotheses, either by Levenshtein distance or timing, are only needed by some metrics so we create them
    # lazily here.
    levenshtein_aligned_hypothesis_segments = None
    time_aligned_hypothesis_segments = None

    results = OrderedDict()

    for metric in args.metrics:
        if metric in results:
            continue  # specified multiple times by the user

        # When using existing parallel segments there will always be a <eob> word match in the end, don't count it.
        # On the other hand, if hypothesis gets aligned to reference a match is not guaranteed, so count it.
        score_break_at_segment_end = False

        full_metric_name = metric
        hypothesis_segments_to_use = hypothesis_segments

        if metric.startswith("AS-"):
            # "AS" stands for automatic segmentation, in particular re-segmentation of the hypothesis using
            # the Levenshtein alignment to the reference.
            # AS-WER and AS-BLEU were introduced by Matusov et al. https://aclanthology.org/2005.iwslt-1.19.pdf
            if levenshtein_aligned_hypothesis_segments is None:
                levenshtein_aligned_hypothesis_segments = levenshtein_align_hypothesis_to_reference(
                    hypothesis=hypothesis_segments, reference=reference_segments)

            hypothesis_segments_to_use = levenshtein_aligned_hypothesis_segments
            metric = metric[len("AS-"):]
            score_break_at_segment_end = True

        elif metric.startswith("t-"):
            # "t" stands for timed. Subtitle timings will be used to re-segment the hypothesis to match the reference
            # segments. t-BLEU was introduced by Cherry et al.
            # https://www.isca-speech.org/archive/pdfs/interspeech_2021/cherry21_interspeech.pdf
            if time_aligned_hypothesis_segments is None:
                time_aligned_hypothesis_segments = time_align_hypothesis_to_reference(
                    hypothesis=hypothesis_segments, reference=reference_segments)

            hypothesis_segments_to_use = time_aligned_hypothesis_segments
            metric = metric[len("t-"):]
            score_break_at_segment_end = True

        elif metric != "SubER" and len(hypothesis_segments_to_use) != len(reference_segments):
            raise ValueError(f"Metric '{metric}' assumes same number of segments in hypothesis and reference, but got "
                             f"{len(hypothesis_segments)} hypothesis and {len(reference_segments)} "
                             f"reference segments.")

        if metric == "SubER":
            metric_score = calculate_SubER(hypothesis=hypothesis_segments_to_use, reference=reference_segments)

        elif metric.startswith("WER"):
            metric_score = calculate_word_error_rate(
                hypothesis=hypothesis_segments_to_use, reference=reference_segments, metric=metric,
                score_break_at_segment_end=score_break_at_segment_end)

        elif metric.startswith("CER"):
            metric_score = calculate_character_error_rate(
                hypothesis=hypothesis_segments_to_use, reference=reference_segments, metric=metric)

        else:
            metric_score = calculate_sacrebleu_metric(
                hypothesis=hypothesis_segments_to_use, reference=reference_segments, metric=metric,
                score_break_at_segment_end=score_break_at_segment_end)

        results[full_metric_name] = metric_score

    json_results = json.dumps(results, indent=4)
    print(json_results)


def check_metrics(metrics):
    allowed_metrics = {
        # Our proposed metric:
        "SubER",
        # Established ASR and MT metrics, requiring aligned hypothesis-references segments:
        "WER", "CER", "BLEU", "TER", "chrF",
        # Cased and punctuated variants of the above:
        "WER-cased", "CER-cased",
        # Segmentation-aware variants of the above that include line breaks as tokens:
        "WER-seg", "BLEU-seg", "TER-seg",
        # Same as "TER-seg" but all words replaced by a mask token,
        # proposed by Karakanta et al. https://aclanthology.org/2020.iwslt-1.26.pdf
        "TER-br",
        # With an "AS-" prefix, the metric is computed after Levenshtein alignment of hypothesis and reference:
        "AS-WER", "AS-CER", "AS-BLEU", "AS-TER", "AS-chrF", "AS-WER-cased", "AS-CER-cased", "AS-WER-seg",
        "AS-BLEU-seg", "AS-TER-seg", "AS-TER-br",
        # With an "t-" prefix, the metric is computed after time alignment of hypothesis and reference:
        "t-WER", "t-CER", "t-BLEU", "t-TER", "t-chrF", "t-WER-cased", "t-CER-cased", "t-WER-seg", "t-BLEU-seg",
        "t-TER-seg", "t-TER-br"}

    invalid_metrics = list(sorted(set(metrics) - allowed_metrics))
    if invalid_metrics:
        raise ValueError(f"Invalid metric(s): {' '.join(invalid_metrics)}")


def check_file_formats(hypothesis_format, reference_format, metrics):
    is_plain_input = (hypothesis_format == "plain" or reference_format == "plain")
    for metric in metrics:
        if ((metric == "SubER" or metric.startswith("t-")) and is_plain_input):
            raise ValueError(f"Metric '{metric}' requires timing information and can only be computed on SRT "
                             f"files (both hypothesis and reference).")


def read_input_file(file_name, file_format):
    if file_format == "SRT":
        file_reader = SRTFileReader(file_name)
    elif file_format == "plain":
        file_reader = PlainFileReader(file_name)
    else:
        raise ValueError(f"Unknown file format: {file_format}")

    file_content = file_reader.read()
    return file_content


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse

from suber.file_readers import SRTFileReader
from suber.sentence_segmentation import resegment_based_on_punctuation
from suber.utilities import segment_to_string


def parse_arguments():
    parser = argparse.ArgumentParser(description="Extracts plain text from SRT files.")
    parser.add_argument("-i", "--input-file", required=True, help="The input SRT file.")
    parser.add_argument("-o", "--output-file", required=True, help="The plain output file.")
    parser.add_argument("-s", "--sentence-segmentation", action="store_true",
                        help="If enabled, output sentences instead of subtitle segments.")

    return parser.parse_args()


def main():
    args = parse_arguments()

    segments = SRTFileReader(args.input_file).read()

    if args.sentence_segmentation:
        segments = resegment_based_on_punctuation(segments)

    with open(args.output_file, "w") as output_file_object:
        for segment in segments:
            output_file_object.write(segment_to_string(segment, include_line_breaks=True) + '\n')


if __name__ == "__main__":
    main()



from enum import Enum
from dataclasses import dataclass
from typing import List


class LineBreak(Enum):
    NONE = 0
    END_OF_LINE = 1  # represented as '<eol>' in plain text files
    END_OF_BLOCK = 2  # represented as '<eob>' in plain text files


@dataclass
class Word:
    string: str
    line_break: LineBreak = LineBreak.NONE  # the line break after the word, if any


@dataclass(unsafe_hash=True)
# Needs to be hashable for cached edit distance in lib_ter.py, but 'approximate_word_time' is currently set after
# creation within SRTFileReader, so cannot set frozen=True.
# TODO: find clean solution
class TimedWord(Word):
    subtitle_start_time: float = None
    subtitle_end_time: float = None
    approximate_word_time: float = None  # usually interpolated from subtitle start and end time; for t-BLEU calculation


@dataclass
class Segment:
    word_list: List[Word]


@dataclass
class Subtitle(Segment):
    index: int
    start_time: float
    end_time: float



import string
from typing import List

from suber.data_types import Subtitle, TimedWord, LineBreak
from suber.constants import END_OF_BLOCK_SYMBOL, END_OF_LINE_SYMBOL
from suber.metrics import lib_ter


def calculate_SubER(hypothesis: List[Subtitle], reference: List[Subtitle]) -> float:
    """
    Main function to caculate the SubER score. We use a modified version of 'lib_ter.py' from sacrebleu for the
    underlying TER implementation. We altered the algorithm by adding a time-overlap condition for word alignments
    and by disallowing word alignments between real words and break tokens.
    """
    total_num_edits = 0
    total_reference_length = 0

    for part in _get_independent_parts(hypothesis, reference):
        hypothesis_part, reference_part = part

        num_edits, reference_length = _calculate_num_edits_for_part(hypothesis_part, reference_part)

        total_num_edits += num_edits
        total_reference_length += reference_length

    if total_reference_length:
        SubER_score = (total_num_edits / total_reference_length) * 100

    elif not total_num_edits:
        SubER_score = 0.0
    else:
        SubER_score = 100.0

    return round(SubER_score, 3)


def _calculate_num_edits_for_part(hypothesis_part: List[Subtitle], reference_part: List[Subtitle]):
    """
    Returns number of edits (word or break edits and shifts) and the total number of reference tokens (words + breaks)
    for the current part.
    """
    all_hypothesis_words = [word for segment in hypothesis_part for word in segment.word_list]
    all_reference_words = [word for segment in reference_part for word in segment.word_list]

    # Although casing and punctuation are important aspects of subtitle quality, we observe higher correlation with
    # human post edit effort when normalizing the words.
    all_hypothesis_words = _normalize_words(all_hypothesis_words)
    all_reference_words = _normalize_words(all_reference_words)

    all_hypothesis_words = _add_breaks_as_words(all_hypothesis_words)
    all_reference_words = _add_breaks_as_words(all_reference_words)

    num_edits, reference_length = lib_ter.translation_edit_rate(
        all_hypothesis_words, all_reference_words)

    assert reference_length == len(all_reference_words)

    return num_edits, reference_length


def _add_breaks_as_words(words: List[TimedWord]) -> List[TimedWord]:
    """
    Converts breaks from being an attribute of the previous Word to being a separate Word in the list. Needed because
    TER algorithm should handle breaks as normal tokens.
    """
    output_words = []
    for word in words:
        output_words.append(
            TimedWord(
                string=word.string,
                line_break=LineBreak.NONE,
                subtitle_start_time=word.subtitle_start_time,
                subtitle_end_time=word.subtitle_end_time,
                approximate_word_time=word.approximate_word_time))

        if word.line_break is not LineBreak.NONE:
            output_words.append(
                TimedWord(
                    string=END_OF_LINE_SYMBOL if word.line_break is LineBreak.END_OF_LINE else END_OF_BLOCK_SYMBOL,
                    line_break=LineBreak.NONE,
                    subtitle_start_time=word.subtitle_start_time,
                    subtitle_end_time=word.subtitle_end_time,
                    approximate_word_time=word.approximate_word_time))

    return output_words


remove_punctuation_table = str.maketrans('', '', string.punctuation)


def _normalize_words(words: List[TimedWord]) -> List[TimedWord]:
    """
    Lower-cases Words and removes punctuation.
    """
    output_words = []
    for word in words:
        normalized_string = word.string.lower()
        normalized_string_without_punctuation = normalized_string.translate(remove_punctuation_table)
        normalized_string_without_punctuation = normalized_string_without_punctuation.replace('…', '')

        if normalized_string_without_punctuation:  # keep tokens that are purely punctuation
            normalized_string = normalized_string_without_punctuation

        output_words.append(
            TimedWord(
                string=normalized_string,
                line_break=word.line_break,
                subtitle_start_time=word.subtitle_start_time,
                subtitle_end_time=word.subtitle_end_time,
                approximate_word_time=word.approximate_word_time))

    return output_words


def _get_independent_parts(hypothesis: List[Subtitle], reference: List[Subtitle]):
    """
    SubER by definition does not require parallel hypothesis-reference segments. We nevertheless split the subtitle file
    content into parts at positions in time where there is no subtitle in both hypothesis and reference. This makes
    calculation more efficient as Levenshtein distances are computed on shorter sequences, while not changing the
    metric score.

    Note, that in the worst case there are no such split points. In practice, this is unrealistic and subtitle files
    are usually limited to a few hours of speech, such that the current SubER calculation should be efficient enough.

    This function yields Tuple[List[Subtitle],List[Subtitle]] containing the hypothesis and reference subtitles for each
    part.
    """
    hypothesis_part = []
    reference_part = []

    # We sweep the time axis from low to high and handle hypothesis and reference subtitles as soon as we reach them.
    hypothesis_subtitle_index = 0  # index of hypothesis subtitle to handle next
    reference_subtitle_index = 0  # index of reference subtitle to handle next
    latest_observed_time = - float('inf')  # highest time observed so far (end time of a previously handled subtitle)

    while hypothesis_subtitle_index < len(hypothesis) or reference_subtitle_index < len(reference):
        if (hypothesis_subtitle_index < len(hypothesis) and (
                reference_subtitle_index == len(reference) or
                hypothesis[hypothesis_subtitle_index].start_time < reference[reference_subtitle_index].start_time)):
            # We found the next subtitle on the time axis, it is from the hypothesis.

            if ((hypothesis_part or reference_part)
                    and hypothesis[hypothesis_subtitle_index].start_time >= latest_observed_time):
                # The subtitle starts after the latest observed time, meaning there is a gap where no subtitle exists.
                # This concludes the current part, yield it.
                yield (hypothesis_part, reference_part)
                hypothesis_part, reference_part = [], []

            hypothesis_part.append(hypothesis[hypothesis_subtitle_index])
            latest_observed_time = max(latest_observed_time, hypothesis[hypothesis_subtitle_index].end_time)
            hypothesis_subtitle_index += 1

        else:  # Next subtitle to handle is from the reference.
            if ((hypothesis_part or reference_part)
                    and reference[reference_subtitle_index].start_time >= latest_observed_time):
                # The subtitle starts after the latest observed time, meaning there is a gap where no subtitle exists.
                # This concludes the current part, yield it.
                yield (hypothesis_part, reference_part)
                hypothesis_part, reference_part = [], []

            reference_part.append(reference[reference_subtitle_index])
            latest_observed_time = max(latest_observed_time, reference[reference_subtitle_index].end_time)
            reference_subtitle_index += 1

    assert hypothesis_subtitle_index == len(hypothesis) and reference_subtitle_index == len(reference)
    if hypothesis_part or reference_part:
        yield (hypothesis_part, reference_part)



import re
import datetime
import numpy


from suber.file_readers.file_reader_base import FileReaderBase
from suber.data_types import LineBreak, TimedWord, Subtitle


class SRTFormatError(Exception):
    pass


class SRTFileReader(FileReaderBase):
    allowed_time_formats = {
        "iso": r"\d+:\d+:\d+\.\d+",
        "iso_with_comma": r"\d+:\d+:\d+,\d+",
        "seconds": r"^\d+(\.\d+)?$",
    }

    def _parse_lines(self, file_object):
        subtitles = []

        subtitle_index = None
        start_time, end_time = None, None
        word_list = None

        for line in file_object:
            line = line.strip()

            if subtitle_index is None:
                # We expect this line to be the subtitle index. Additional empty line is okay too.
                if line:
                    try:
                        subtitle_index = int(line.replace("\ufeff", ""))
                    except ValueError as e:
                        raise SRTFormatError(f"Tried to read subtitle index from '{line}' but failed.") from e

            elif start_time is None:
                # We expect this line to be the time string. Additional empty line is okay too.
                if line:
                    start_time, end_time = self._parse_time_stamp(line)
                    if end_time < start_time:
                        raise SRTFormatError(f"End time {end_time} is before start time {start_time}.")

                    assert word_list is None
                    word_list = []  # start collecting words

            elif line:
                # We expect this line to contain subtitle text.
                assert subtitle_index is not None
                assert start_time is not None
                assert end_time is not None

                # We don't consider formatting tags <i>, <b>, etc. in the evaluation.
                # TODO: maybe we want this regex to cover more cases
                line = re.sub('</?[^>]>', '', line)

                word_list.extend([
                    TimedWord(
                        string=word,
                        subtitle_start_time=start_time,
                        subtitle_end_time=end_time)
                    for word in line.split()])

                if word_list:
                    word_list[-1].line_break = LineBreak.END_OF_LINE

            else:
                # This is an empty line after lines of subtitle text which ends the current subtitle.
                assert word_list is not None
                assert start_time is not None
                assert end_time is not None

                if word_list:  # might be an empty subtitle
                    word_list[-1].line_break = LineBreak.END_OF_BLOCK

                    self._set_approximate_word_times(word_list, start_time, end_time)

                subtitles.append(
                    Subtitle(word_list=word_list, index=subtitle_index, start_time=start_time, end_time=end_time))

                subtitle_index = None
                start_time, end_time = None, None
                word_list = None

        if word_list is not None:
            # handle last subtitle
            assert subtitle_index is not None
            assert start_time is not None and end_time is not None

            if word_list:  # might be an empty subtitle
                word_list[-1].line_break = LineBreak.END_OF_BLOCK

                self._set_approximate_word_times(word_list, start_time, end_time)

            subtitles.append(
                Subtitle(word_list=word_list, index=subtitle_index, start_time=start_time, end_time=end_time))

        return subtitles

    @classmethod
    def _set_approximate_word_times(cls, word_list, start_time, end_time):
        """
        Linearly interpolates word times from the subtitle start and end time as described in
        https://www.isca-speech.org/archive/pdfs/interspeech_2021/cherry21_interspeech.pdf
        """
        # Remove small margin to guarantee the first and last word will always be counted as within the subtitle.
        epsilon = 1e-8
        start_time = start_time + epsilon
        end_time = end_time - epsilon

        num_words = len(word_list)
        duration = end_time - start_time
        assert duration >= 0

        approximate_word_times = numpy.linspace(start=start_time, stop=end_time, num=num_words)
        for word_time, word in zip(approximate_word_times, word_list):
            word.approximate_word_time = word_time

    @classmethod
    def _parse_time_stamp(cls, time_stamp):
        time_stamp_tokens = time_stamp.split()
        if len(time_stamp_tokens) != 3 or not time_stamp_tokens[1].endswith("->"):
            raise SRTFormatError(f"Could not parse subtitle times from '{time_stamp}'.")

        start_time = cls._seconds_from_time_code(time_stamp_tokens[0])
        end_time = cls._seconds_from_time_code(time_stamp_tokens[2])

        return start_time, end_time

    @classmethod
    def _seconds_from_time_code(cls, time_code):
        detected_time_format = None
        for time_format_name, time_format_regex in cls.allowed_time_formats.items():
            if re.match(time_format_regex, time_code):
                detected_time_format = time_format_name
                break

        if not detected_time_format:
            raise SRTFormatError(f"Could not detect format of time code '{time_code}'.")

        try:
            if detected_time_format == "seconds":
                seconds = float(time_code)
            else:
                assert detected_time_format in ["iso", "iso_with_comma"]
                datetime_format_string = "%H:%M:%S,%f" if detected_time_format == "iso_with_comma" else "%H:%M:%S.%f"

                datetime_object = datetime.datetime.strptime(time_code, datetime_format_string)
                # Set year to 1 (default is 1900) to set this time in relation to datetime.datetime.min.
                datetime_object = datetime_object.replace(year=1)

                seconds = (datetime_object - datetime.datetime.min).total_seconds()
        except Exception as e:
            raise SRTFormatError(f"Could not convert '{time_code}' to seconds. "
                                 f"Tried to read it as format '{detected_time_format}'.") from e

        return seconds



from suber.file_readers.file_reader_base import FileReaderBase
from suber.data_types import LineBreak, Word, Segment
from suber.constants import END_OF_LINE_SYMBOL, END_OF_BLOCK_SYMBOL


class PlainFileReader(FileReaderBase):
    def _parse_lines(self, file_object):
        segments = []

        is_first_line = True
        for line in file_object:
            if is_first_line:
                if line.startswith('\ufeff'):
                    line = line[len('\ufeff'):]  # remove byte order mark (BOM)
                is_first_line = False
            words = line.split()

            word_list = []
            for word in words:
                if word in (END_OF_LINE_SYMBOL, END_OF_BLOCK_SYMBOL):
                    if not word_list:
                        continue  # ignore line break symbol at the start of the line
                    else:
                        word_list[-1].line_break = (
                            LineBreak.END_OF_BLOCK if word == END_OF_BLOCK_SYMBOL else LineBreak.END_OF_LINE)
                else:
                    word_list.append(Word(string=word))

            segments.append(Segment(word_list=word_list))

        return segments
